# include <stdio.h>
# include <math.h>

# include "apm.h"
# include "apmUtil.h"
# include "apmSpecial.h"
# include "converse.h"
# include "bounding.h"
# include "rows.h"
# include "map.h"
# include "pi.h"

# define  DEBUG     NO 

# define  USE_CR    YES 
# define  USE_LL    YES
# define  USE_QT    YES 
# define  USE_RIGOR YES 
# define  USE_SHIFT NO 

# define  FF_CYCLS  1 
# define  QS_TO_RS  5	/* The ratio of quick steps to rigorous steps
			   to be used in determining how long 
			   quick_try should go.
			*/
# define  MAX_SUM   PI

# define  setLclFurthest( n )	( ((((n)/QS_TO_RS)+3) > furthest) \
				? furthest : ((n/QS_TO_RS)+3) )

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
	Varaibles shared via converse.h 
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int		quick_tries, tries, Rtries, max_steps, max_NTsteps ;
int		HermSuccess, LLSuccess, ll_used[3], most_cuts ;

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
	Varaibles private to this file.
	
    The functions below manipulate copies of the data passed
    to them and local  copies are kept in Prisms and RPrisms
    gotten with the conjuring functions by initFollowing(). 
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

static Prism	*workPriz[2] ;

static double	b_buf[DF_SQ], *b_ptrs[DF_SQ] ;
static double	parmbuf[2*N_PARMS], coordbuf[2*TWO_DF] ;

static Xtnd_pt	xpt_a, xpt_b ;

/* Some APM variables needed for orbit following and  slope watching. */
static RPrism	*Rwork[2] ;

static APM		f_scratch, Rdenom ;
static APM		Rsum, Rmax_sum ;
static APM		Rtrace_ll, RmaxBlam_ll, RminBlam_ll ;
static double	trace_ll, maxBlam_ll, minBlam_ll ;

/*
	The variables declared below don't really need to 
	be bounded objects (they did in an earlier version of the code),
	but the .ub in their uses makes the code easier to understand.
*/
static Bdd_dbl	   b_trace, minBlam, maxBlam, leastLam, slope ;
static Bdd_apm	   Rb_trace, RminBlam, RmaxBlam, RleastLam, Rslope ;

static int	is_first_trial = YES ;
static int	local_furth ;

/* Functions private to this file */
static int quick_try(RPrism *priz, double LLmax);
/* +++++++++++++++++++++++++++++++++ */

void 
prepare_trial (RPrism *priz)
{
    int   j ; 

    if( areNewParms( priz ) == YES ) {

/*
	Unless this is the very first prism, 
	record the center point - it will be moved by 
	setHermStart() and setLLStart() and will neeed to be
	restored to its correct value.
*/
	if( is_first_trial == NO ) { 
	    for ( j=0 ; j < DEG_FREE ; j++ ) {
	    	apmAssign( xpt_a.z.u[j], priz->center->z.u[j] ) ;
	    	apmAssign( xpt_a.z.v[j], priz->center->z.v[j] ) ;
	    }
	}

	setHermStart( priz ) ;
	setCone( priz ) ;
# if USE_LL
	setLLStart( priz ) ;
	setLLbounds( priz ) ;
# endif
# if USE_SHIFT
	shiftStart( priz ) ;
# endif
/*
	     Unless this is the very first trial, restore the 
	     correct value of the centerpoint before evaluating
	     the initial estimates for the slope and least eigenvalue.
*/
	if( is_first_trial == YES )
	     is_first_trial = NO ;
	else {
	    for ( j=0 ; j < DEG_FREE ; j++ ) {
	    	apmAssign( priz->center->z.u[j], xpt_a.z.u[j]  ) ;
	    	apmAssign( priz->center->z.v[j], xpt_a.z.v[j]  ) ;
	    }
	}

	setSlopes( priz ) ;

# if USE_LL
	setLeastLam( priz ) ; 
# else
	firstLeastLam = 1.0 ;
	minLeastLam = 0.5 ;
	dbltoapm( RfirstLeastLam, BASE, firstLeastLam ) ;
	dbltoapm( RminLeastLam, BASE, minLeastLam ) ;
# endif
    }
}
/* +++++++++++++++++++++++++++++++++ */

int 
try_prism (RPrism *initial_priz, Prism *final_priz, int *nsteps)
/*
     Decides whether a prism of initial data may 
     contain any invariant Lagrangian tori.  It checks
     the trace condition at each point along an orbit.
     The check proceeds in two stages. First, we just
     compute images of the initial point (the center
     of the initial prism) and check the trace condition
     as we go along.  If this method suggests that 
     no torus exists, we do the same calculation with
     more elaborate, prism-bounding techniques.  If the
     first check goes to "local_furth" images without
     getting a definite answer we quit.  The prism-bounding
     part quits when either the prism grows too large
     (How large is too large is determined by the
     global variable big_prism.), or when "local_furth"
     images have been considered.
*/
{
    int       quick_verdict, count ;
    Prism     *priz, *priz_prime, *temp_priz ;
    double    denom, best_ll() ;
/*
	Get a quick verdict.  If it's only MAYBE the longer
	calculation will certainly return MAYBE too.
*/
    isNewPrism = YES ;
    priz = workPriz[0] ;
    priz_prime = workPriz[1] ;
    copyRPrism( priz, initial_priz ) ;

# if USE_QT
    local_furth = setLclFurthest( priz->n_cuts ) ;
    quick_verdict = quick_try( initial_priz, firstLeastLam ) ;
# else
    local_furth = furthest ;
    quick_verdict = NO_TORI ;
# endif

    if( quick_verdict ==  MAYBE ) {
	*nsteps = 0 ;
	copyRPrism( final_priz, initial_priz ) ;
	return( MAYBE ) ;
    }
/*
	If the quick verdict was NO_TORI, press on and try a more
	careful prism-bounding sort of thing.
*/
    else {
	tries++ ;
# if DEBUG
	print_perf() ;
# endif

	fatten = fxed_form ;
	row_sums = ff_rows ;

	*nsteps = count = 1 ;
	slope.ub = first_slope ;
	leastLam.ub = firstLeastLam ;
	if( (slope.ub < min_slope) || (leastLam.ub < minLeastLam) ) {
	    copyPrism( final_priz, priz ) ;
	    return( NO_TORI ) ;
	}

	while( big_Prism( priz ) == NO ) {
	    count++ ;
/*
		Bounds useful for all criteria
*/
	    global_bounds( priz ) ;
	    bound_btrace( &b_trace, priz ) ;

# if USE_LL
	/*
			mrm's condition
			Calculated three ways, two here and
			one more after the new slope.
	*/
	    bd_Blams( &minBlam, &maxBlam, &b_trace ) ;

	    maxBlam_ll = maxBlam.ub - 1.0 / leastLam.ub ;

	    denom = slope.ub - sumTinyLams ;
	    if( denom > 0.0 ) 
	        minBlam_ll = minBlam.ub - 1.0 / denom ;
	    else
		minBlam_ll = 0.0 ;
# endif

	/*
		Herman's Condition
	*/
	    slope.ub = b_trace.ub - (DF_SQ / slope.ub) ;

# if USE_LL
	/*
		Calculate the third _ll and choose the
		best one.
	*/
	    trace_ll = slope.ub / DEG_FREE ;

	    leastLam.ub = best_ll( maxBlam_ll, 
				   minBlam_ll, trace_ll ) ;

# endif
# if DEBUG
	    printf( "try_prism : slope_ub = %.6f, ll_ub = %.6f \n",
							slope.ub,
							leastLam.ub ) ;
# endif
	    if( (slope.ub < min_slope) || (leastLam.ub < minLeastLam) ) {
		*nsteps = count ;
		if( count > max_steps )
		    max_steps = count ;

		copyPrism( final_priz, priz ) ;
		return( NO_TORI ) ;
	    }
	    else {
		if( count == furthest )
		    break ;

		prismatic_image( priz_prime, priz ) ;
		m_swap( priz, priz_prime, temp_priz ) ;
	    }
# if USE_CR
	    if( count > FF_CYCLS )  {
	        fatten = col_rotor ;
	        row_sums = cr_rows ;
	    }
# endif
	}
    }
  
    *nsteps = count ;
    if( count > max_steps )
	max_steps = count ;

    copyPrism( final_priz, priz ) ;
    return( MAYBE ) ;
}
/* +++++++++++++++++++++++++++++++++++++++ */

int 
quick_try (RPrism *priz, double LLmax)
{
    int	     count, k ;
    double   slope, b_trace, *b_diagPt, calc_bML(), b_ML ;
    Xtnd_pt  *x, *x_prime, *temp ;

    quick_tries++ ;
# if DEBUG
	print_perf() ;
# endif
/*
	Set things up and copy the variables.
*/
    x = &xpt_a ;
    x_prime = &xpt_b ;
    for( k=0 ; k < N_PARMS ; k++ ) 
	x->p[k] = apmtodbl( priz->center->p[k] ) ;

    for( k=0 ; k < DEG_FREE ; k++ ) {
	x->z.u[k] = apmtodbl( priz->center->z.u[k] ) ;
	x->z.v[k] = apmtodbl( priz->center->z.v[k] ) ;
    }

/*
	Do many iterations, stopping at "local_furth" or when the
	trace condition is violated.
*/
    slope = first_slope ;
    count = 1 ;
    while( count < local_furth )  {
/*
		Check the slope.
*/
	if( (slope < min_slope) || (LLmax < 0.0) ) {
# if DEBUG
	    printf( "Quick_try gets no tori in %d iterations. \n", count ) ;
# endif
	    return( NO_TORI ) ;
	}

	else {
/*
		Act with the map; obtain the interesting parts 
		of its jacobian.
*/
	    image( x_prime, x ) ;
	    find_beta( b_ptrs, x ) ;

/*
		Find the new trace.
*/
	    b_trace  = 0.0 ;
	    b_diagPt = b_buf ;
	    for( k=0 ; k < DEG_FREE ; k++ ) {
		b_trace += *b_diagPt ;
		b_diagPt += DEG_FREE  + 1 ;
	    }

# if USE_LL
	    b_ML = calc_bML( b_buf, b_trace ) ;
	    LLmax = b_ML - 1.0 / LLmax ; 
# endif
	    slope = b_trace - (DF_SQ / slope) ;
	    m_swap( x, x_prime, temp ) ;
	}
	count++ ;
    }

# if DEBUG
    printf( "Quick_try gets maybe after %d iterations. \n", count ) ;
# endif

    return( MAYBE ) ;
}
/* +++++++++++++++++++++++++++++++++ */

void 
initFollowing (void)
{
/*
	Set up the correct connections between the various
	static variables in this file.
*/
    int	      j, all_well ;

    all_well = YES ;
/*
	Set up the working prisms.
*/
    workPriz[0] = conjurePrism() ;
    workPriz[1] = conjurePrism() ;
    if( (workPriz[0] == NULL) || (workPriz[1] == NULL) ) 
	all_well = NO ;

/*
	Set up the APM stuff
*/
    f_scratch = apmNew( BASE ) ;
    Rdenom = apmNew( BASE ) ;

    Rtrace_ll = apmNew( BASE ) ;
    RminBlam_ll = apmNew( BASE ) ;
    RmaxBlam_ll = apmNew( BASE ) ;

    newBapm( Rslope, BASE ) ;
    newBapm( Rb_trace, BASE ) ;
    newBapm( RminBlam, BASE ) ;
    newBapm( RmaxBlam, BASE ) ;  
    newBapm( RleastLam, BASE ) ;

# if (USE_LL == NO)
    apmAssignLong( RleastLam.ub, 1L, 0, BASE ) ;
    apmAssignLong( RleastLam.lb, 1L, 0, BASE ) ;
# endif

    Rsum = apmNew( BASE ) ;
    Rmax_sum = apmNew( BASE ) ;
    dbltoapm( Rmax_sum, BASE, MAX_SUM ) ;
    
    Rwork[0] = conjureRPrism() ;
    Rwork[1] = conjureRPrism() ;
    if( (Rwork[0] == NULL) || (Rwork[1] == NULL) )
	all_well = NO ;

/*
	Set up the extended points - they're used by 
	quick_test(), and are pointed to by the
	"center" attributes of the working prisms. 
*/
    xpt_a.z.u = coordbuf ;
    xpt_a.z.v = coordbuf + DEG_FREE ;
    xpt_a.p   = parmbuf ;

    xpt_b.z.u = coordbuf + TWO_DF ;
    xpt_b.z.v = coordbuf + TWO_DF + DEG_FREE ;
    xpt_b.p   = parmbuf + N_PARMS ;
/*
	Set up pointers to the matrix which receives the 
	changeable parts of the jacobian; the one called 
	"beta" in most of my notes.  
*/
    for( j=0 ; j < (sizeof( b_buf ) / sizeof( double )) ; j++ )
	b_ptrs[j] = &b_buf[j] ;

/*
	Initialize various performance data.
*/
    most_cuts = 0 ;
    max_steps = max_NTsteps = 1 ;
    HermSuccess = LLSuccess = 0 ;
    tries = Rtries = quick_tries = 0 ;
    ll_used[0] = ll_used[1] = ll_used[2] = 0 ;

    if( all_well == YES ) 
	return ;
    else {
	fprintf( stderr, 
		 "initFollowing : Could not prepare for pursuit. Died. \n" ) ;
	cease() ;
    }
}
/* ++++++++++++++++++++++++ */

int 
Rtry_prism (RPrism *initial_priz, Prism *final_priz, int *nsteps)
/*
     Rigorously decides whether a prism of initial data may 
     contain any invariant Lagrangian tori, an APM version of
     the routine tryPrism() above.
*/
{
    int		count ;
    RPrism      *priz, *priz_prime, *temp_priz ;

    Rtries++ ;
    priz = Rwork[0] ;
    priz_prime = Rwork[1] ;
# if DEBUG
	print_perf() ;
# endif

/*
	Note that Rtry_prism() does not call setSlopes, setStart or
	setCone.  All that should have been done with a call to 
	prepare_trial().
*/

    isNewPrism = YES ;
    RcopyRPrism( priz, initial_priz ) ;

    fatten = Rfxed_form ;
    row_sums = Rff_rows ;

    *nsteps = count = 1 ;
    apmAssign( Rslope.ub, Rfirst_slope ) ;
    apmAssign( RleastLam.ub, RfirstLeastLam ) ;
    if( apmCompare(Rslope.ub, Rmin_slope) == -1 ) {
	HermSuccess++ ;
	copyRPrism( final_priz, priz ) ;
	return( NO_TORI ) ;
    }
    if( apmCompare(RleastLam.ub, RminLeastLam) == -1 ) {
	LLSuccess++ ;
	copyRPrism( final_priz, priz ) ;
	return( NO_TORI ) ;
    }

# if (USE_RIGOR == NO)
    copyRPrism( final_priz, priz ) ;
    return( NO_TORI ) ;
# endif

    while( big_RPrism( priz ) == NO ) {
/*
		Check the slope.
*/
	count++ ;
/*
		Calculate some bounds useful for both criteria.
*/
	Rglobal_bounds( priz ) ;
	Rbound_btrace( &Rb_trace, priz ) ;

# if USE_LL
		/* mrm's condition  */
	Rbd_Blams( &RminBlam, &RmaxBlam, &Rb_trace ) ;

	apmDivide( f_scratch, precision, (APM)NULL, one,
					 RleastLam.ub ) ;
	apmSubtract( RmaxBlam_ll, RmaxBlam.ub, f_scratch ) ;


	apmSubtract( Rdenom, Rslope.ub, RsumTinyLams ) ;
	if( apmCompare( Rdenom, zero) > 0 ) {
	    apmDivide( f_scratch, precision, (APM) NULL, one, Rdenom ) ;
	    apmSubtract( RminBlam_ll, RminBlam.ub, f_scratch ) ;
	}
	else 
	    apmAssign( RminBlam_ll, zero ) ;
# endif

		/* Herman's condition */
	apmDivide( f_scratch, precision, (APM) NULL, Rdf_sq, Rslope.ub ) ;
	apmSubtract( Rslope.ub, Rb_trace.ub, f_scratch ) ;

# if USE_LL
	apmDivide( Rtrace_ll, precision, (APM)NULL, Rslope.ub, Rdf ) ;

	Rbest_ll( RleastLam.ub, RmaxBlam_ll,
		    RminBlam_ll, Rtrace_ll ) ;

# endif

# if DEBUG
	printf("Rslope = %.6f, RleastLam = %.6f\n", apmtodbl( Rslope.ub ),
						    apmtodbl( RleastLam.ub ) ) ;
# endif

/* 		
		Do some truncations to speed things up
*/
# if USE_LL
	apmTruncate( RleastLam.ub, precision ) ;
# endif
	apmTruncate( Rslope.ub, precision ) ;

        if( apmCompare(Rslope.ub, Rmin_slope) == -1 ) {
	    *nsteps = count ;
	    if( count > max_NTsteps ) 
		max_NTsteps = count ;

	    HermSuccess++ ;
	    copyRPrism( final_priz, priz ) ;
	    return( NO_TORI ) ;
   	 }
	else if( apmCompare(RleastLam.ub, RminLeastLam) == -1 ) {
	    *nsteps = count ;
	    if( count > max_NTsteps ) 
		max_NTsteps = count ;

	    LLSuccess++ ;
	    copyRPrism( final_priz, priz ) ;
	    return( NO_TORI ) ;
	}
	else {
	    if( count == furthest )
		break ;

	    Rprismatic_image( priz_prime, priz ) ;

	    m_swap( priz, priz_prime, temp_priz ) ;
	}
# if USE_CR
	if( count > FF_CYCLS )  {
	    fatten = Rcol_rotor ;
	    row_sums = Rcr_rows ;
	}
# endif
    }
  
    *nsteps = count ;
    copyRPrism( final_priz, priz ) ;
    return( MAYBE ) ;
}
/* +++++++++++++++++++++++++++++++++++++++ */

int 
big_Prism (Prism *Priz)
{
    double   sum, *rpt, *end_mat, *end_row ;

    end_mat = Priz->matrix + MAT_SZ ;
    for( rpt = Priz->matrix ; rpt < end_mat ; ) {
	sum = 0.0 ;
	for( end_row = rpt + MAT_DIM ; rpt < end_row ; rpt++ )
	    sum += fabs( *rpt ) ;

	if( sum > MAX_SUM ) 
	    return( YES ) ;
    }

    return( NO ) ;
}
/* +++++++++++++++++++++++++++++++++++++++ */

int 
big_RPrism (RPrism *Priz)
{
    APM   *Rrpt, *Rend_mat, *Rend_row ;

    Rend_mat = Priz->matrix + MAT_SZ ;
    for( Rrpt = Priz->matrix ; Rrpt < Rend_mat ; ) {
	apmAssignLong( Rsum, 0L, 0, BASE ) ;
	for( Rend_row = Rrpt + MAT_DIM ; Rrpt < Rend_row ; Rrpt++ )
	    apmCalc( Rsum, Rsum, *Rrpt, APM_ABS, APM_ADD, NULL ) ;

	if( apmCompare( Rsum, Rmax_sum) == 1 ) 
	    return( YES ) ;
    }

    return( NO ) ;
}
/* ++++++++++++++++++++++ */

void 
print_perf (void)
{
    printf( "\n" ) ;
    printf( "quick_tries = %d, tries = %d, Rtries = %d. \n",
				quick_tries, tries, Rtries ) ;
    printf( "max_steps = %d, max_NTsteps = %d \n", max_steps,
						max_NTsteps ) ;
    printf( "maxBlam_ll = %d, minBlam_ll = %d, trace_ll = %d \n",
					ll_used[0], ll_used[1],
						ll_used[2] ) ;
    printf( "\n" ) ;
}
/* ++++++++++++++++++++++++++++++++++++ */

double 
best_ll (double minBlam_ll, double maxBlam_ll, double trace_ll)
{
    int	    chosen ;
    double  best ;

    best = maxBlam_ll ;
    chosen = 0 ;
    if( best > minBlam_ll ) {
	best = minBlam_ll ;
	chosen = 1 ;
    }

    if( best > trace_ll ) {
	best = trace_ll ;
	chosen = 2 ;
    }

    ll_used[chosen]++ ;
    return( best ) ;
}
/* +++++++++++++++++++++++ */

void 
Rbest_ll (APM best, APM minBlam_ll, APM maxBlam_ll, APM trace_ll)
{
    apmAssign( best, maxBlam_ll ) ;
    if( apmCompare( best, minBlam_ll ) == 1 ) 
	apmAssign( best, minBlam_ll ) ;

    if( apmCompare( best, trace_ll ) == 1 ) 
	apmAssign( best, trace_ll ) ;
}
