# include <stdio.h>
# include <stdlib.h>
# include <math.h>

# include "apm.h"
# include "apmUtil.h"
# include "apmSpecial.h"
# include "converse.h"

# define DEBUG		    NO 
# if DEBUG
# include "apmPrint.h"
# endif

# define OOM_DF		    1 
# define MAX_RECUR	    2
# define DFLT_XDP	    5

# define Rm_swap(x,y,t)	    (apmAssign(t, x), apmAssign(x, y), \
								apmAssign(y, t) )
/*
	The Numerical Recipes Gauss-Jordan matrix inverter as adaptaed
	for a converse KAM code.
	I have removed the dimension arguments n and m and replaced 
	them with TWO_DF and 1.  I have also changed all the floats 
	into doubles and replaced some automatically allocated
	arrays with arrays of fixed dimension.  Finally, I have 
	replaced the error handling code with some of my own.

	Rgauss, the rigorous version, also does a host of checks to
	guarantee that the inverse it produces, when multiplied by 
	the original matrix, a, gives something equal to the 
	identity to the accuracy specified by the global variable,
	"precision".
*/

/* Functions private to this file */
static void copyRmat(APM **copy, APM **mat);
static int choosePrecis(APM **mat);


/* Variables private to this file */
static int	extra_dp, last_inv_dp ;
static int	inv_depth ;	/*	Used to make sure that we don't keep trying
 			   				to invert singular matrices by using 
 			   				ever increasing precision. 	*/

static APM	a_abs, Rbig, Rdum, Rpivinv, Rtemp ;
static APM	Rrow_max, Rcol_max, Rmat_min, Rmat_max ;
static APM	*Rmat[TWO_DF], Rmat_block[4*DF_SQ] ;
static APM	Rdiv_err, Rrow_err, Rinv_err, Rtotal_err, Rpiv_err ;
/* ++++++++++++++++++++++++++++++++ */

void 
gauss (double **a)
{
	int indxc[TWO_DF],indxr[TWO_DF],ipiv[TWO_DF];
	int i,icol,irow,j,k,l,ll;
	double big,dum,pivinv,temp;

	for( j=0 ; j < TWO_DF ; j++ ) {
		ipiv[j] = 0 ;
		indxr[j] = 0 ;
		indxc[j] = 0 ;
	}

	for (j=0;j<TWO_DF;j++) ipiv[j]=0;
	for (i=0;i<TWO_DF;i++) {
		big=0.0;
		for (j=0;j<TWO_DF;j++)
			if (ipiv[j] != 1)
				for (k=0;k<TWO_DF;k++) {
					if (ipiv[k] == 0) {
						if (fabs(a[j][k]) >= big) {
							big=fabs(a[j][k]);
							irow=j;
							icol=k;
						}
					} else if (ipiv[k] > 1) {
					fprintf( stderr,
					"Singular matrix in gauss. Died.\n" ) ;
					cease() ;
					}
				}
		++(ipiv[icol]);
		if (irow != icol) {
			for (l=0;l<TWO_DF;l++) 
				m_swap(a[irow][l],a[icol][l],temp) ;
		}
		indxr[i]=irow;
		indxc[i]=icol;
		if (a[icol][icol] == 0.0) {
		    fprintf( stderr,"Singular matrix in gauss(). Died.\n" ) ;
		    cease() ;
		}
		pivinv=1.0/a[icol][icol];
		a[icol][icol]=1.0;
		for (l=0;l<TWO_DF;l++) a[icol][l] *= pivinv;
		for (ll=0;ll<TWO_DF;ll++)
			if (ll != icol) {
				dum=a[ll][icol];
				a[ll][icol]=0.0;
				for (l=0;l<TWO_DF;l++) a[ll][l] -= a[icol][l]*dum;
			}
	}
	for (l=(TWO_DF-1);l>=0;l--) {
		if (indxr[l] != indxc[l])
			for (k=0;k<TWO_DF;k++)
				m_swap(a[k][indxr[l]],a[k][indxc[l]],temp);
	}
}
/* ++++++++++++++++++++++++++++ */

void 
initGauss (void)
{
    int   j, k ;
    APM   *mpt ;

    inv_depth = 0 ;
    extra_dp = 0 ;

    Rbig = apmNew( BASE ) ;
    Rdum = apmNew( BASE ) ;
    a_abs = apmNew( BASE ) ;
    Rtemp = apmNew( BASE ) ;
    Rpivinv = apmNew( BASE ) ;
    Rinv_err = apmNew( BASE ) ;
    Rrow_err = apmNew( BASE ) ;
    Rpiv_err = apmNew( BASE ) ;
    Rdiv_err = apmNew( BASE ) ;
    Rrow_max = apmNew( BASE ) ;
    Rcol_max = apmNew( BASE ) ;
    Rmat_min = apmNew( BASE ) ;
    Rmat_max = apmNew( BASE ) ;
    Rtotal_err = apmNew( BASE ) ;

    mpt = Rmat_block ;
    for( j=0 ; j < TWO_DF ; j++ ) {
	Rmat[j] = mpt ;
	for( k=0 ; k < TWO_DF ; k++ ) 
	    *mpt++ = apmNew( BASE ) ;
    }
}
/* ++++++++++++++++++++++++++++ */

void 
Rgauss (APM **a)
{
	int indxc[TWO_DF],indxr[TWO_DF],ipiv[TWO_DF];
	int i,icol,irow,j,k,l,ll;
	int inv_dp, err_dp ;
# if DEBUG
	char  buf[BUF_SZ] ;
# endif

	if( ++inv_depth > MAX_RECUR ) {
	    fprintf( stderr, "Singular matrix in Rgauss. Died. \n" ) ;
	    cease() ;
	}

	for( j=0 ; j < TWO_DF ; j++ ) {
	    ipiv[j] = 0 ;
	    indxr[j] = 0 ;
	    indxc[j] = 0 ;
	}
/*
	    If this is the attempt to invert a,
	    copy the matrix in case of a loss of precision. 
	    Also, choose
	    the precision to which to do the inversion calculations.
*/
	if( inv_depth == 1 ) {
	    copyRmat( Rmat, a ) ;
	    inv_dp =  choosePrecis( a ) ;
	}
	else {
	    if( extra_dp == 0 ) 
		inv_dp = last_inv_dp + DFLT_XDP ;
	    else
		inv_dp = last_inv_dp + extra_dp ;
	}
	last_inv_dp = inv_dp ;		
# if DEBUG
	fprintf( stderr, 
		 "Doing an inversion, carrying %d digits. \n", inv_dp ) ;
# endif

/*
	Initialize the error propagation stuff.
*/
	apmAssignLong( Rdiv_err, 1L, -inv_dp, BASE ) ;
	apmAssignLong( Rinv_err, 0L, 0, BASE ) ;
	apmAssign( Rpiv_err, Rinv_err ) ;

	for (i=0;i<TWO_DF;i++) {
	    apmAssignLong( Rbig, 0L, 0, BASE ) ;
	    for (j=0;j<TWO_DF;j++) {
		if (ipiv[j] != 1) {
		    for (k=0;k<TWO_DF;k++) {
			if (ipiv[k] == 0) {
			    apmAbsoluteValue( a_abs, a[j][k] ) ;
			    if( apmCompare(a_abs, Rbig) != -1 ) {
				apmAssign( Rbig, a_abs ) ;
				irow=j;
				icol=k;
			    }
			} 
			else if (ipiv[k] > 1) {
			    fprintf( stderr,
				"Singular matrix in gauss. Died.\n" ) ;
			    cease() ;
			}
		    }
		}
	    }

	    ++(ipiv[icol]);
	    if(irow != icol) {
		for (l=0;l<TWO_DF;l++) 
		    Rm_swap(a[irow][l],a[icol][l],Rtemp) ;
	    }

	    indxr[i]=irow;
	    indxc[i]=icol;
/*
		Check that the pivot interval does not
		contain zero.  If it does, restart the
		calculation and carry more decimal places.
*/
	    apmCalc( Rtemp, a[icol][icol], APM_ABS,
			    Rinv_err, APM_SUB, NULL ) ;
	    if( apmCompare( Rtemp, zero ) != 1 ) { 
		copyRmat( a, Rmat ) ;
# if DEBUG
		fprintf( stderr, "Bad pivot interval. \n" ) ;
		fflush( stderr ) ;
# endif
		Rgauss( a ) ;
		return ;
	    }
/*
		Get the new pivot error.  It is here that we face
		the possibility of catastrophic loss of precision.
*/
	    apmDivide( Rpiv_err, inv_dp, (APM)NULL, Rinv_err, Rtemp ) ;
	    apmCalc( Rpiv_err, Rpiv_err, Rdiv_err, Rdiv_err, 
					 APM_ADD, APM_ADD, NULL ) ;
	    apmDivide(Rpivinv,inv_dp,(APM)NULL,one,a[icol][icol]) ;
	    apmAssignLong( a[icol][icol], 1L, 0, BASE ) ;
# if DEBUG
	    apmConvert( buf, BUF_SZ, inv_dp, DO_ROUND, LEFT_JUST, Rpiv_err ) ;
	    fprintf( stderr, "piv_error : %d, %s \n", 
			     apmLogBd( Rpiv_err ), buf  ) ;
	    fprintf( stderr, "div_error : %d \n", apmLogBd( Rdiv_err ) ) ;
# endif
	  
	    apmAssignLong( Rrow_max, 0L, 0, BASE ) ;
	    for (l=0;l<TWO_DF;l++) { 
		if( l != icol ) {
		    apmAbsoluteValue( Rtemp, a[icol][l] ) ;
		    if( apmCompare( Rtemp, Rrow_max ) < 0 )
			apmAssign( Rrow_max, Rtemp ) ;
		}

		apmCalc(a[icol][l], a[icol][l], Rpivinv,APM_MUL,NULL) ;
	    }

/*
		Get a bound on the size of the errors in the elements
		of the pivot row.
*/
	    apmCalc( Rrow_err,  Rinv_err, Rpivinv, APM_MUL,
				Rrow_max, Rinv_err, APM_ADD,
				Rpiv_err, APM_MUL, APM_ADD, NULL ) ;
# if DEBUG
	    fprintf( stderr, "row_error : %d\n", apmLogBd( Rrow_err ) ) ;
# endif

	    apmAssignLong( Rcol_max, 0L, 0, BASE ) ;
	    for (ll=0;ll<TWO_DF;ll++) {
		if (ll != icol) {
		    apmAssign( Rdum, a[ll][icol] ) ;
		    apmAbsoluteValue( Rtemp, Rdum ) ;
		    if( apmCompare( Rtemp, Rcol_max ) == 1 )
			apmAssign( Rcol_max, Rtemp ) ;

		    apmAssignLong( a[ll][icol], 0L, 0, BASE ) ;
		    for (l=0;l<TWO_DF;l++) 
			apmCalc( a[ll][l], a[ll][l], a[icol][l], Rdum,
					   APM_MUL, APM_SUB, NULL ) ;
		}
	    }

/*
		Calculate the new upper bound on errors in the matrix.
*/
	    apmCalc( Rinv_err,  Rrow_max, Rrow_err, APM_ADD, 
				Rinv_err, APM_MUL,
				Rcol_max, Rrow_err, APM_MUL,
				Rinv_err, APM_ADD, 
				APM_ADD, APM_ADD, NULL ) ;
/*
	Add an extra Rdiv_err to Rinv_err and truncate everything.
	This will probably speed the calculation considerably.
*/
	    apmCalc( Rinv_err, Rinv_err, Rdiv_err, APM_ADD, NULL ) ;
# if DEBUG
	    fprintf( stderr, "inv_error : %d\n\n", 
				apmLogBd( Rinv_err ) ) ;
# endif

	    apmTruncate( Rinv_err, inv_dp ) ;
	    for( l = 0 ; l < TWO_DF ; l++ )
	    	for( ll=0 ; ll < TWO_DF ; ll++ ) 
		    apmTruncate( a[l][ll], inv_dp ) ;
	}


	for (l=(TWO_DF-1);l>=0;l--) {
	    if (indxr[l] != indxc[l])
		for (k=0;k<TWO_DF;k++)
		    Rm_swap(a[k][indxr[l]],a[k][indxc[l]],Rtemp);
	}
/*
	Check the overall size of the error.
	If it is too big, set extra_dp and try again.
*/
	err_dp = -(apmLogBd( Rinv_err ) + OOM_DF) ;
	if( err_dp < precision ) { 
		extra_dp = precision - err_dp + 2 ;
		copyRmat( a, Rmat ) ;
# if DEBUG
		fprintf( stderr, "Poor answer : errdp = %d \n", err_dp ) ;
		fflush(stderr) ;
# endif
		Rgauss( a ) ;
		return ;
	}

/*
	Tidy up.
	If we reach this line, all is well, the inversion is
	good to the desired precision, so all we want to do is 
	restore the recurrsive variables to their initial state.
*/
    inv_depth = 0 ;
    extra_dp = 0 ;
    return ;
}
/* +++++++++++++++++++++++++++++++++ */

void 
copyRmat (APM **copy, APM **mat)
{
    int  j, k ;

    for( j=0 ; j < TWO_DF ; j++ ) 
	for( k=0 ; k < TWO_DF ; k++ )
	    apmAssign( copy[j][k], mat[j][k] ) ;
}
/* ++++++++++++++++++++++++++++++++++ */

int 
choosePrecis (APM **mat)
{
    APM  *mpt, *end_mat ;
    int  oom_min, oom_max, oom_err, oom_twos ;

/*
	Find the minimum and maximum entries of the matrix.
	If none of the entries has absolute value bigger than 
	one, use one as the maximum; this ensures that the 
	resulting inverse will have entries good to at least
	"precision" decimal places.
*/
    mpt = mat[0] ;
    apmAssignLong( Rmat_min, 0L, 0, BASE ) ;
    apmAssignLong( Rmat_max, 1L, 0, BASE ) ;

    for( end_mat = mpt + (TWO_DF*TWO_DF) ; mpt < end_mat ; mpt++ ) {
	apmAbsoluteValue( Rtemp, *mpt ) ;
	if( apmCompare( Rmat_min, Rtemp ) > 0 )
		apmAssign( Rmat_min, Rtemp ) ;
	else if( apmCompare( Rmat_max, Rtemp ) < 0 )
	    apmAssign( Rmat_max, Rtemp ) ;
    }

/*
	Do a basic estimate of the number of digits one must carry 
	to get an answer whose precision is as good as the code
	requires.
	    First find the orders of magnitude ("oom"'s) of various things.
*/
    oom_max = apmLogBd( Rmat_max ) ;
    oom_twos = (TWO_DF / 3) ;

    oom_err = oom_twos + OOM_DF + (2 * TWO_DF + 1) * abs( oom_max )  ;

    if( oom_err < 0 )
		return( precision ) ;
    else
		return( precision + oom_err ) ;
}
