/*
	Functions to apply the extended Froeschle map and to
	calculate its jacobian.  Each function has a rigorous
	and a non-rigorous form; the former always has a name
	beginning with a "R".

	The functions in this file are quite specific - 
	they pertain to maps  of the form

	(p,u,v) -> (p',u',v')

	p' = p
	u' = v
	v' = 2v - u -dV(v)

	where u, v, u' and v' are all in 2d Euclidean space,
	p is an element of a space of parameters and
	    V(v) = -a * sin( v[0] ) + -b * sin( v[1] ) +
		   -c * sin( v[0] + v[1] )

	The parameters a, b, and c are always passed through
	an array called "parms" with 

	a = parms[0], b = parms[1], c = parms[2].
*/
# include <stdio.h>
# include <math.h>

# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"

/* Variables shared through converse.h */
APM		RDeriv[MAT_SZ], *Rbeta_ptrs[DF_SQ], *Rgamma_ptrs[DF_SQ] ;
double	Deriv[MAT_SZ], *beta_ptrs[DF_SQ], *gamma_ptrs[DF_SQ] ;

/* Functions private to this file */
static void setRv_sum(RXtnd_pt *x);
static void find_gamma(double *g_block[], Xtnd_pt *x);
static void find_Rgamma(APM *g_block[], RXtnd_pt *x);
extern void initJacobian(double *jac);
extern void initRjacobian(APM *jac);

/* Variables private to this file */
static APM		Rmixing_term, Rv_sum, map_scratch ;
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++
	setRv_sum() : saves a little time by evaluating 
	this sum only once.
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
setRv_sum (RXtnd_pt *x)
{
    apmAdd( Rv_sum, x->z.v[0], x->z.v[1] ) ;
}
/* +++++++++++++++++++++++++++++++++++++++++ 
	image() and Rimage() 
+++++++++++++++++++++++++++++++++++++++++ */

void 
image (Xtnd_pt *x_prime, Xtnd_pt *x)
/*
	Finds the image, x_prime,  of a delay-embedded point, x.
	The parameters of the map are in the parameter-space point
	called "parms".
*/
{
    double   mixing_term, *x_pt, *xp_pt, *last_x ;
    Parm_pt  parms ;

    parms = x->p ;
    x_pt = x->p ;
    xp_pt = x_prime->p ;
    for( last_x = x_pt + N_PARMS ; x_pt < last_x ; x_pt++ )
	*xp_pt++ = *x_pt ;

	/* Because of the way delay embedding works, 
	   the first member of x_prime is the same as 
	   the second member of x .	
	*/

    x_pt = x->z.v ;
    xp_pt = x_prime->z.u ;
    for( last_x = x_pt + DEG_FREE ; x_pt < last_x ; x_pt++ )
	*xp_pt++ = *x_pt ;

	/*  Do up the actual map.  One could 
	    write a version of image() which worked for
	    any standard-type symplectic map; it would 
	    rely on another function, perturb(), to 
	    completely define the map.  Instead we
	    incorporate the perturbation to the 
	    generating function right into our map -
	    we hope to save a little time.
	*/
    mixing_term = parms[2] * cos( (x->z.v[0] + x->z.v[1]) ) ;
    x_prime->z.v[0] = 2.0 * x->z.v[0] - x->z.u[0]   +
		    parms[0] * cos( x->z.v[0] ) +
		    mixing_term			;
    x_prime->z.v[1] = 2.0 * x->z.v[1] - x->z.u[1]   +
		    parms[1] * cos( x->z.v[1] ) +
		    mixing_term			;
}
/* +++++++++++++++++++++++++++++++++++++++++ */

void 
Rimage (RXtnd_pt *x_prime, RXtnd_pt *x)
/*
	Finds the image, x_prime,  of a delay-embedded point, x.
	The parameters of the map are in the parameter-space point
	called "parms".
*/
{
    APM		*x_pt, *xp_pt, *last_x ;
    RParm_pt	parms ;

    parms = x->p ;
    x_pt = x->p ;
    xp_pt = x_prime->p ;
    for( last_x = x_pt + N_PARMS ; x_pt < last_x ; x_pt++ )
	apmAssign( *xp_pt++, *x_pt ) ;

	/* Because of the way delay embedding works, 
	   the first member of x_prime is the same as 
	   the second member of x .	
	*/

    x_pt = x->z.v ;
    xp_pt = x_prime->z.u ;
    for( last_x = x_pt + DEG_FREE ; x_pt < last_x ; x_pt++ )
	apmAssign( *xp_pt++, *x_pt ) ;

	/*  Do up the actual map.  One could 
	    write a version of image() which worked for
	    any standard-type symplectic map; it would 
	    rely on another function, perturb(), to 
	    completely define the map.  Instead we
	    incorporate the perturbation to the 
	    generating function right into our map -
	    we hope to save a little time.
	*/
    apmAdd( Rv_sum, x->z.v[0], x->z.v[1] ) ;
    apmCos( map_scratch, Rv_sum ) ;
    apmMultiply( Rmixing_term, map_scratch, parms[2] ) ;

    apmCos( map_scratch, x->z.v[0] ) ;
    apmCalc( x_prime->z.v[0], two, x->z.v[0], APM_MUL, 
				 x->z.u[0], APM_SUB, 
				 parms[0], map_scratch, APM_MUL,
				 Rmixing_term, APM_ADD,
				 APM_ADD, NULL			) ;

    apmCos( map_scratch, x->z.v[1] ) ;
    apmCalc( x_prime->z.v[1], two, x->z.v[1], APM_MUL, 
				 x->z.u[1], APM_SUB, 
				 parms[1], map_scratch, APM_MUL,
				 Rmixing_term, APM_ADD,
				 APM_ADD, NULL			) ;
}
/* +++++++++++++++++++++++++++++++++++++++++++
	find_beta() and find_Rbeta()

	In the interest of speed, we provide functions which only
	calculate those parts of the jacobian that actually
	depend on parms and (u,v).  The other parts are 
	assumed to have been correctly set by a call to
	initJacobian() or initRjacobian(), both of which
	may be found below.
+++++++++++++++++++++++++++++++++++++++++++ */

void 
find_beta (double *b_block[], Xtnd_pt *x)
/*
	The components of b_block should be pointers to those 
	elements of the jacobian which depend on the coordinates
	of the extended point.
								*/
{
    double   mixing_term ;

    mixing_term = x->p[2] * sin( x->z.v[0] + x->z.v[1] ) ;

    *b_block[0] = 2.0 - x->p[0] * sin( x->z.v[0] ) - mixing_term ;
    *b_block[1] = -mixing_term ;
    *b_block[2] = -mixing_term ;
    *b_block[3] = 2.0 - x->p[1] * sin( x->z.v[1] ) - mixing_term ;
}
/* +++++++++++++++++++++ */

void 
find_Rbeta (APM *b_block[], RXtnd_pt *x)
{
    apmAdd( Rv_sum, x->z.v[0], x->z.v[1] ) ;
    apmSin( map_scratch, Rv_sum ) ;
    apmMultiply( Rmixing_term, x->p[2], map_scratch ) ;

    apmSin( map_scratch, x->z.v[0] ) ;
    apmCalc( *b_block[0], x->p[0], map_scratch, APM_MUL,
			  two, APM_SWAP, APM_SUB,
			  Rmixing_term, APM_SUB, NULL ) ;

    apmNegate( *b_block[1], Rmixing_term ) ;
    apmNegate( *b_block[2], Rmixing_term ) ;

    apmSin( map_scratch, x->z.v[1] ) ;
    apmCalc( *b_block[3], x->p[1], map_scratch, APM_MUL,
			  two, APM_SWAP, APM_SUB,
			  Rmixing_term, APM_SUB, NULL ) ;
}
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++ 

    gamma() and Rgamma() : calculate the dependence of 
    v' on the parameters.  Even as the functions
    above, gamma() and Rgamma() change only those components
    pointed to by elements of a block of pointers.  
+++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
find_gamma (double *g_block[], Xtnd_pt *x)
{
     double  mixing_term ;

     mixing_term = cos( x->z.v[0] + x->z.v[1] ) ;

     *g_block[0] = cos( x->z.v[0] ) ;
     *g_block[1] = mixing_term ;
     *g_block[2] = cos( x->z.v[1] ) ;
     *g_block[3] = mixing_term ;
}
/* ++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
find_Rgamma (APM *g_block[], RXtnd_pt *x)
{
    apmAdd( Rv_sum, x->z.v[0], x->z.v[1] ) ;
    apmCos( Rmixing_term, Rv_sum ) ;

    apmCos( *g_block[0], x->z.v[0] ) ;
    apmAssign( *g_block[1], Rmixing_term ) ;
    apmCos( *g_block[2], x->z.v[1] ) ;
    apmAssign( *g_block[3], Rmixing_term ) ;
}
/* ++++++++++++++++++++++++++++++++++++++++ */

void 
initJacobian (double *jac)
{
    int	    j ;
    double  *end_jac, *jpt ;

    end_jac = jac + MAT_SZ ;			/* Set all the entries */
    for( jpt=jac ; jpt < end_jac ; jpt++ )	/* to zero.		*/
	*jpt = 0.0 ;

/*	Put the identity in the (p,p) position. */
    jpt = jac ;
    for( j=0 ; j < N_PARMS ; j++ ) {
	*jpt = 1.0 ;
	jpt += MAT_DIM + 1 ;
    }

/*	Put the identity in the (u,v) position. */
    jpt = jac + STAID_LEN + N_PARMS + DEG_FREE ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	*jpt = 1.0 ;
	jpt += MAT_DIM + 1 ;
    }

/*	Put -1 times the identity in the (v,u) position. */
    jpt = jac + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	*jpt = -1.0 ;
	jpt += MAT_DIM + 1 ;
    }
}
/* ++++++++++++++++++++++++++++++++++++++++ */

void 
initRjacobian ( 
/*
	Set the constant parts of a jacobian matrix
*/
    APM *jac
)
{
    int	    j ;
    APM     *end_jac, *jpt ;

/*	
	If the array of APM's called jac has not yet been 
	initialized, do that first.
*/
    if( jac[0] == NULL) {
		end_jac = jac + MAT_SZ ;
		for( jpt=jac ; jpt < end_jac ; jpt++ ) {
			*jpt = apmNew( BASE ) ;
		}
	}

	end_jac = jac + MAT_SZ ;			/* Set all the entries */
	for( jpt=jac ; jpt < end_jac ; jpt++ )	/* to zero.		*/
	apmAssignLong( *jpt, 0L, 0, BASE ) ;

	/*	Put the identity in the (p,p) position. */
	jpt = jac ;
	for( j=0 ; j < N_PARMS ; j++ ) {
		apmAssignLong( *jpt, 1L, 0, BASE ) ;
		jpt += MAT_DIM + 1 ;
		}

	/*	Put the identity in the (u,v) position. */
	jpt = jac + STAID_LEN + N_PARMS + DEG_FREE ;
	for( j=0 ; j < DEG_FREE ; j++ ) {
		apmAssignLong( *jpt, 1L, 0, BASE ) ;
		jpt += MAT_DIM + 1 ;
		}

	/*	Put -1 times the identity in the (v,u) position. */
	jpt = jac + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
	for( j=0 ; j < DEG_FREE ; j++ ) {
		apmAssignLong( *jpt, -1L, 0, BASE ) ;
		jpt += MAT_DIM + 1 ;
    }
}
/* +++++++++++++++++++++ */

void 
initMap (void)
{
/*
    This function depends in detail on the choice of map.
*/
    beta_ptrs[0] = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) +
					N_PARMS  + DEG_FREE ;
    beta_ptrs[1] = beta_ptrs[0] + 1 ;
    beta_ptrs[2] = beta_ptrs[0] + MAT_DIM ;
    beta_ptrs[3] = beta_ptrs[2] + 1 ;

    gamma_ptrs[0] = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    gamma_ptrs[1] = gamma_ptrs[0] + 2 ;
    gamma_ptrs[2] = gamma_ptrs[0] + MAT_DIM + 1 ;
    gamma_ptrs[3] = gamma_ptrs[1] + MAT_DIM ;

/*
	APM stuff
*/
    Rbeta_ptrs[0] = RDeriv + STAID_LEN + (DEG_FREE * MAT_DIM) +
						N_PARMS + DEG_FREE ;
    Rbeta_ptrs[1] = Rbeta_ptrs[0] + 1 ;
    Rbeta_ptrs[2] = Rbeta_ptrs[0] + MAT_DIM ;
    Rbeta_ptrs[3] = Rbeta_ptrs[2] + 1 ;

    Rgamma_ptrs[0] = RDeriv + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Rgamma_ptrs[1] = Rgamma_ptrs[0] + 2 ;
    Rgamma_ptrs[2] = Rgamma_ptrs[0] + MAT_DIM + 1 ;
    Rgamma_ptrs[3] = Rgamma_ptrs[1] + MAT_DIM ;

	RDeriv[0] = NULL ;
    initJacobian( Deriv ) ;
    initRjacobian( RDeriv ) ;
/*
	Further APM stuff - constants and scratch variables.
*/
    Rv_sum = apmNew( BASE ) ;
    map_scratch = apmNew( BASE ) ;
    Rmixing_term = apmNew( BASE ) ;
}
/* +++++++++++++++++++++++ */

void 
jacobian (Xtnd_pt *xpt)
{
    find_beta( beta_ptrs, xpt ) ;
    find_gamma( gamma_ptrs, xpt ) ;
}
/* +++++++++++++++++++++++ */

void 
Rjacobian (RXtnd_pt *xpt)
{
    find_Rbeta( Rbeta_ptrs, xpt ) ;
    find_Rgamma( Rgamma_ptrs, xpt ) ;
}
