# include <stdio.h>
# include <stdlib.h>
# include <math.h>
# include "apm.h"
# include "converse.h"
# include "tree.h"

/*  
	Functions to handle prisms.  We allocate biggish blocks of
	prisms with a call to prismAlloc(), initialize them, and then
	parcel them out to other functions with the conjurePrism()
	function. Other functions may also return prisms they no
	longer need - this they do with releasePrism().	

*/

/* Functions private to this file */
static Prism *PrismAlloc(int n);
static RPrism *RPrismAlloc(int n);
static Xtnd_pt *xptAlloc(int n);
static RXtnd_pt *RxptAlloc(int n);
static PrismLeaf *LeafAlloc(int n);

/* Variables private to this file */
static Prism     *idle_prisms = NULL ;
static RPrism    *idle_RPrisms = NULL ;
static PrismLeaf *idle_leaves = NULL ;
/* +++++++++++++++++++++++++++++++++++++++++++++++++++ */

Prism *
conjurePrism (void)
{
    Prism   *result ;

	/* If need be, get some new prisms */
    if( idle_prisms == NULL ) 
	idle_prisms = PrismAlloc( 2 * depth + 2 ) ;

	/* Take the next idle prism. */
    result = idle_prisms ;
    if( result != NULL ) {
	idle_prisms = result->next ;
	result->next = NULL ;
    }
    else 
	fprintf( stderr, "Trouble conjuring prisms. \n" ) ;

    return( result ) ;
}
/* +++++++++++++++++++++++++++ */

void 
releasePrism (Prism *ppt)
{
    if( ppt != NULL ) {
		ppt->next = idle_prisms ;
		idle_prisms = ppt ;
    }
}
/* +++++++++++++++++++++++++++ */

Prism *
PrismAlloc ( 
/*
     Allocate enough memory for n prisms, and 
     for all the stuff that goes into them.
*/
    int n
)
{
    int		j, cut_strlen, cut_blksz  ;
    char	*strings, *spt ;
    Prism	*result, *ppt, *end_chain ;
    double	*mats, *mpt, *end_mat ;
    Xtnd_pt 	*centers, *cpt ; 

    cut_strlen = depth + 2 ;
    cut_blksz  = MAT_DIM * cut_strlen ;

    mats = (double *) malloc( n * MAT_SZ * sizeof(double) ) ;
    result = (Prism *) malloc( n * sizeof(Prism) ) ;
    strings = malloc( n * cut_blksz ) ;
    centers = xptAlloc( n ) ;
    if( (mats == NULL) || (result == NULL) ||
	(strings == NULL) || (centers == NULL) ) {
	    fprintf( stderr, "Trouble allocating prisms. \n" ) ;

	    m_free( (char *) mats ) ;
	    m_free( (char *) result ) ;
	    m_free( (char *) strings ) ;
	    m_free( (char *) centers ) ;

	    result = NULL ;
    }
    else {
	mpt = mats ;
	cpt = centers ;
	spt = strings ;

	end_chain = result + n ;
	for( ppt = result ; ppt < end_chain ; ppt++ ) {
	    ppt->in_torus = UNTRIED ;
	    ppt->n_cuts = 0 ;

	    for(  j=0 ; j < MAT_DIM ; j++ ) { 
		spt[0] = '.' ;
		spt[1] = '\0' ;

		ppt->cuts[j] = spt ;
		spt += cut_strlen ;
	    }

	    ppt->matrix = mpt ;
	    for( end_mat =  mpt + MAT_SZ ; mpt < end_mat ; mpt++ )
		*mpt = 0.0 ;

	    ppt->center = cpt++ ;

	    ppt->next = ppt + 1 ;
	}

	result[n-1].next = (Prism *) NULL ;
    }

    return( result ) ;
}
/* +++++++++++++++++++++++++++++ */
	
RPrism *
conjureRPrism (void)
{
    RPrism   *result ;

	/* If need be, get some new prisms */
    if( idle_RPrisms == NULL ) 
	idle_RPrisms = RPrismAlloc( 2 * depth + 2) ;

	/* Take the next idle prism. */
    result = idle_RPrisms ;
    if( result != NULL ) {
	idle_RPrisms = result->next ;
	result->next = NULL ;
    } 
    else 
	fprintf( stderr, "Trouble conjuring APM prisms. \n" ) ;

    return( result ) ;
}
/* +++++++++++++++++++++++++++ */

void 
releaseRPrism (RPrism *ppt)
{
    if( ppt != NULL ) {
		ppt->next = idle_RPrisms ;
		idle_RPrisms = ppt ;
    }
}
/* +++++++++++++++++++++++++++ */

RPrism *
RPrismAlloc ( 
/*
     Allocate enough memory for n RPrisms, and 
     for all the stuff that goes into them.
*/
    int n
)
{
    int		j, cut_strlen, cut_blksz  ;
    APM		*mats, *mpt, *end_mat ;
    char	*strings, *spt ;
    RPrism	*result, *ppt, *end_chain ;
    RXtnd_pt	*centers, *cpt ; 

    cut_strlen = depth + 2 ;
    cut_blksz  = MAT_DIM * cut_strlen ;

    mats = (APM *) malloc( n * MAT_SZ * sizeof(APM) ) ;
    result = (RPrism *) malloc( n * sizeof(RPrism) ) ;
    strings = malloc( n * cut_blksz ) ;
    centers = RxptAlloc( n ) ;
    if( (mats == NULL) || (result == NULL) ||
	(strings == NULL) || (centers == NULL) ) {
	    fprintf( stderr, "Trouble allocating prisms. \n" ) ;

	    m_free( (char *) mats ) ;
	    m_free( (char *) result ) ;
	    m_free( (char *) strings ) ;
	    m_free( (char *) centers ) ;

	    result = NULL ;
    }
    else {
	mpt = mats ;
	cpt = centers ;
	spt = strings ;

	end_chain = result + n ;
	for( ppt = result ; ppt < end_chain ; ppt++ ) {
	    ppt->in_torus = UNTRIED ;
	    ppt->n_cuts = 0 ;

	    for(  j=0 ; j < MAT_DIM ; j++ ) { 
		spt[0] = '.' ;
		spt[1] = '\0' ;

		ppt->cuts[j] = spt ;
		spt += cut_strlen ;
	    }

	    ppt->matrix = mpt ;
	    for( end_mat = mpt + MAT_SZ ; mpt < end_mat ; mpt++ )
		*mpt = apmNew( BASE ) ;

	    ppt->center = cpt++ ;

	    ppt->next = ppt + 1 ;
	}

	result[n-1].next = NULL ;
    }


    return( result ) ;
}
/* ++++++++++++++++++++++ */

Xtnd_pt *
xptAlloc (int n)
{
    double   *coords, *cpt ;
    Xtnd_pt  *result, *xpt, *end_x ;

    coords = (double *) malloc( n * MAT_DIM * sizeof(double) ) ;
    result = (Xtnd_pt *) malloc( n * sizeof(Xtnd_pt) ) ;

    if( (result == NULL) || (coords == NULL) ) {
	m_free( (char *) result ) ;
	m_free( (char *) coords ) ;

	fprintf( stderr, "Trouble allocating extended points. \n" ) ;
	result = NULL ;
    }
    else {
	cpt = coords ;

	end_x = result + n ;
	for( xpt = result ; xpt < end_x ; xpt++ ) {
	    xpt->p = cpt ;
	    xpt->z.u = cpt + N_PARMS ;
	    xpt->z.v = cpt + N_PARMS + DEG_FREE ;

	    cpt += MAT_DIM ;
	}
    }

    return( result ) ;
}
/* ++++++++++++++++++++++++++++++ */

RXtnd_pt *
RxptAlloc (int n)
{
    APM       *coords, *cpt, *end_c ;
    RXtnd_pt  *result, *xpt, *end_x ;

    coords = (APM *) malloc( n * MAT_DIM * sizeof(double) ) ;
    result = (RXtnd_pt *) malloc( n * sizeof(Xtnd_pt) ) ;

    if( (result == NULL) || (coords == NULL) ) {
		m_free( (char *) result ) ;
		m_free( (char *) coords ) ;

		fprintf( stderr, "Trouble allocating extended points. \n" ) ;
		result = NULL ;
    }
    else {
		cpt = coords ;

		end_x = result + n ;
		for( xpt = result ; xpt < end_x ; xpt++ ) {
			xpt->p = cpt ;
			xpt->z.u = cpt + N_PARMS ;
			xpt->z.v = cpt + N_PARMS + DEG_FREE ;

			for( end_c = cpt + MAT_DIM ; cpt < end_c ; cpt++ )
			*cpt = apmNew( BASE ) ;
		}
    }

    return( result ) ;
}
/* ++++++++++++++++++++++++++++++ */

void 
m_free (char *cpt)
{
    if( cpt != NULL ) 
		free( cpt ) ;
}
/* +++++++++++++++++++++++++++++++ */

# if FANCY_TREE
PrismLeaf *
conjureLeaf (void)
{
    PrismLeaf    *result ;

	/* If need be, get some new leaves */
    if( idle_leaves == NULL ) 
		idle_leaves = LeafAlloc( 2 * depth + 2 ) ;

	/* Take the next idle prism. */
    result = idle_leaves ;
    if( result != NULL ) {
		idle_leaves = result->right ;
		result->left = NULL ;
		result->right = NULL ;
    }
    else 
		fprintf( stderr, "Trouble conjuring leaves. \n" ) ;

    return( result ) ;
}
/* +++++++++++++++++++++++++++ */

PrismLeaf *
LeafAlloc ( 
/*
     Allocate enough memory for n leaves, and 
     for all the stuff that goes into them.
*/
    int n
)
{
    int		j, cut_strlen, cut_blksz  ;
    char	*strings, *spt ;
    PrismLeaf   *lpt, *result, *end_chain ; 

    result = (PrismLeaf *) malloc( n * sizeof(PrismLeaf) ) ;

    cut_strlen = depth + 2 ;
    cut_blksz  = MAT_DIM * cut_strlen ;

    strings = malloc( n * cut_blksz ) ;
    if( (strings == NULL) || (result == NULL) ) {
	    fprintf( stderr, "Trouble allocating leaves. \n" ) ;

	    m_free( (char *) strings ) ;
	    m_free( (char *) result ) ;
	    result = NULL ;
    }
    else {
		spt = strings ;

		end_chain = result + n ;
		for( lpt = result ; lpt < end_chain ; lpt++ ) {
			lpt->color = UNTRIED ;

			for(  j=0 ; j < MAT_DIM ; j++ ) { 
			spt[0] = '.' ;
			spt[1] = '\0' ;

			lpt->cuts[j] = spt ;
			spt += cut_strlen ;
			}

			lpt->left = NULL ;
			lpt->right = lpt + 1 ;
		}

		result[n-1].right = NULL ;
    }

    return( result ) ;
}
# endif
