# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>

# include "apm.h"
# include "apmUtil.h"
# include "apmPrint.h"
# include "apmSpecial.h"

# define  DEBUG   NO 

/* Variables shared through apmSpecial.h */
APM    zero, one, two, four, eight, sixteen ;
APM    pi, two_pi, half_pi, threeHalf_pi, eighths_2pi[8] ;

/* Variables private to this file */
static APM    Theta, scratch, xMod2pi, Theta_sq, Answer ;
static APM    *sinCoef, *cosCoef ;
static APM    util_scratch ;
static APM    sinFactrl, cosFactrl, apmOrder ;
static APM    approx[2], diff, ub_diff ;
static int    trig_dp, specialsInit = NO  ;
static int    trig_terms, dp_lost ;
static char   pi_str[] = "3.14159265358979323846243383279502884197169399375" ;
static char   log_buf[BUF_SZ] ; /* BUF_SZ is defined in apmUtil.h */

/* Functions private to this file */
static int reduceArg( APM x ) ;
static void reducedSin( void ) ;
static void reducedCos( void ) ;
static void strip_frac( char *str ) ;
static void setTrigCoef( void ) ;
static void apmFloorString( char *s, int n, APM x ) ;
/* ++++++++++++++++++++++++ */

int initApmSpecials( int base )
{
    int     k ;

	/* Initialize a bunch of APMs.  Theta will be the reduced argument 
	   of a trig function; it will be between zero and pi / 4.  	*/
      
    /* Powers of 2 */
    zero = apmInit( 0L, 0, base ) ;
    one = apmInit( 1L, 0, base ) ;
    two = apmInit( 2L, 0, base ) ;
    four = apmInit( 4L, 0, base ) ;
    eight = apmInit( 8L, 0, base ) ;
    sixteen = apmInit( 16L, 0, base ) ;
    
	/* Thins private to this file */
    diff = apmNew( base );
    Theta = apmNew( base );
    Answer = apmNew( base );
    scratch = apmNew( base );
    ub_diff = apmNew( base );
    xMod2pi = apmNew( base );
    apmOrder = apmNew( base );
    Theta_sq = apmNew( base );
    sinFactrl = apmNew( base );
    cosFactrl = apmNew( base );
    approx[0] = apmNew( base );
    approx[1] = apmNew( base );
    util_scratch = apmNew( base );

	/* Obtain some rational mutiples of pi.  These will be helpful
	   when we go to restrict the domain of the trig functions to 
	   between zero and pi / 4 .					*/

  	/* First initialise them */
    pi = apmNew( base ); 
    two_pi = apmNew( base );
    half_pi = apmNew( base );
    threeHalf_pi = apmNew( base );
    
    for( k=0 ; k < 8 ; k++ )
		eighths_2pi[k] = apmNew( base );

	/* Then set their values */
    apmAssignString( pi, pi_str, 0 ) ;

    apmMultiply( scratch, two, two ) ;
    apmDivide( eighths_2pi[0], (PI_DP+2), (APM)NULL, pi, scratch) ;

    for( k=1 ; k < 8 ; k++ ) 
	apmAdd( eighths_2pi[k], eighths_2pi[0], eighths_2pi[k-1] ) ;

    apmMultiply( two_pi, pi, two ) ;
    apmAssign( half_pi, eighths_2pi[1] ) ;
    apmAssign( threeHalf_pi, eighths_2pi[5] ) ;

    setTrigDp( DFLT_TRIG_DP ) ;

    dp_lost = 0 ;
    specialsInit = YES ;

    return( 1 ) ;
}
/* ++++++++++++++++++++++++++ */

int 
setTrigDp (int dp)
{
    double  j, j_fact, ten_to_dp ;

	/* Check to see that the desired accuracy is compatible
	   with our knowledge of pi.				*/

    if( (dp+2) > PI_DP ) {
	fprintf( stderr, 
	"We don't know pi well enough to achieve the desired accuracy. \n" ) ;
	return( 0 ) ;
    }
    else
	trig_dp = dp+2 ;

	/* Assume the argument is between zero and pi / 4.  How many
	   terms from the Taylor series do we need to include ?     */

    trig_terms = 1 ;
    ten_to_dp = pow( 10.0, (double)dp ) ;
    for( j = 1.0, j_fact = 1.0 ; j_fact < ten_to_dp ; j += 2.0 )  {
	j_fact *= j * (j + 1) ;
	trig_terms++ ;
	if( trig_terms > MAX_TRIG_TERMS ) {
	    fprintf( stderr, "Too many terms required. \n" ) ;
	    return(0) ;
	}
    }

    trig_dp += (int)( ceil( log10((double) trig_terms) ) ) ;
    setTrigCoef() ;
    return( dp ) ;
}
/* +++++++++++++++++++++++++++++++++++++ */

int 
reduceArg (
/*
    Takes x, chops off enough multiples of two_pi to get it 
    into the interval between zero and two_pi.  Checks that we
    haven't lost an unacceptable amount of precision in doing
    this stage of the reduction.  Then  chops off multiples
    of pi/4 to get the argument into the interval between zero and
    pi/4.  Sets Theta equal to the reduced argument and returns
    an integer indicating in which of eight equally spaced intervals
    x (mod two_pi) lay.  If any precision is lost, dp_lost is set 
    to the number of decimal places lost.
									*/
    APM x
)
{
    int   octant ;
    char  qtnt_str[BUF_SZ] ;

	/* Note that we haven't lost any decimal places yet. */
    dp_lost = 0 ;

	/* Whack out many multiples of two_pi. */
    apmDivide( scratch, 3, (APM)NULL, x, two_pi ) ;
    apmFloorString( qtnt_str, BUF_SZ, scratch ) ;
    apmAssignString( scratch, qtnt_str, 0 ) ;
    apmMultiply( Answer, scratch, two_pi ) ;
    apmSubtract( xMod2pi, x, Answer ) ;
    if( apmSign( xMod2pi ) == -1 ) 
	 apmCalc( xMod2pi, xMod2pi, two_pi, APM_ADD, NULL ) ;

    for( octant=0 ;  (octant < 8)  ; octant++ ) {
	 if( apmCompare(xMod2pi, eighths_2pi[octant]) < 0 )
	     break ;
    }

    switch( octant ) {
	case 0 :
	    apmAssign( Theta, xMod2pi ) ;
	    break ;

	case 1 :
	    apmSubtract( Theta, half_pi, xMod2pi ) ;
	    break ;

	case 2 :
	    apmSubtract( Theta, xMod2pi, half_pi ) ;
	    break ;

	case 3 :
	    apmSubtract( Theta, pi, xMod2pi ) ;
	    break ;

	case 4 :
	    apmSubtract( Theta, xMod2pi, pi ) ;
	    break ;

	case 5 :
	    apmSubtract( Theta, threeHalf_pi, xMod2pi ) ;
	    break ;

	case 6 :
	    apmSubtract( Theta, xMod2pi, threeHalf_pi ) ;
	    break ;

	case 7 :
	    apmSubtract( Theta, two_pi, xMod2pi ) ;
	    break ;

	default :
	    break ;
    }

	/* Check for loss of precision */
    if( (PI_DP - strlen(qtnt_str)) < trig_dp )
	dp_lost = trig_dp - PI_DP + strlen(qtnt_str) ;
    else
	dp_lost = 0 ;

    return( octant ) ;
}
/* +++++++++++++++++++++++ */

void reducedSin( void )
/*
    Takes the sine of Theta, puts the result in Answer.
								*/
{
    int  order, dp_to_find, term_num ; 

    apmAssign( Answer, zero ) ;
    apmMultiply( Theta_sq, Theta, Theta ) ;

    term_num = trig_terms - 1 ;
    for( order = ( 2 * trig_terms - 1 ) ; order > 0 ; order -= 2 ) {

	    /* Multiply the old partial sum by Theta squared
	       and add in a new coefficient			*/
	apmMultiply( scratch, Answer, Theta_sq ) ;
	apmAdd( Answer, sinCoef[term_num--], scratch ) ;
	apmTruncate( Answer, trig_dp ) ;
    }

	/*  Multiply by the final factor of Theta,
	    divide by the factorial, and return  */

    if( dp_lost > 0 ) 
	dp_to_find = trig_dp + 1 - dp_lost ;
    else
	dp_to_find = trig_dp + 1 ;

    apmMultiply( scratch, Answer, Theta ) ;
    apmDivide( Answer, dp_to_find, (APM)NULL, scratch, sinFactrl ) ;
    return ;
}
/* ++++++++++++++++++++++++++++++++++++ */

void reducedCos( void )
/*
    Takes the cosine of Theta, puts the result in Answer.
								*/
{
    int  order, dp_to_find, term_num ; 

    apmAssign( Answer, zero ) ;
    apmMultiply( Theta_sq, Theta, Theta ) ;

    term_num = trig_terms - 1 ;
    for( order = ( 2 * trig_terms - 2 ) ; order >= 0 ; order -= 2 ) {

	    /* Multiply the old partial sum by Theta squared
	       and add in a new coefficient			*/
	apmMultiply( scratch, Answer, Theta_sq ) ;
	apmAdd( Answer, cosCoef[term_num--], scratch ) ;

	apmTruncate( Answer, trig_dp ) ;
    }

	/*  Divide by the factorial, 
	    Put the result into Answer, and return  */

    if( dp_lost > 0 ) 
	dp_to_find = trig_dp + 1 - dp_lost ;
    else
	dp_to_find = trig_dp + 1 ;

    apmDivide( scratch, dp_to_find, (APM)NULL, Answer, cosFactrl ) ;
    apmAssign( Answer, scratch ) ;
    return ;
}
/* ++++++++++++++++++++++++++++++++++++ */

void 
apmSin (APM result, APM x)
{
    int   octant ;

    if( specialsInit == NO ) {
	fprintf( stderr, 
		 "apmSin() : Please call initApmSpecials(). \n" ) ;
 	apmAssignLong( result, 0L, 0, 0 ) ;
	apm_errno = APM_EPARM ;
	return ;
    }
    else 
	apm_errno = APM_OK ;

	/* Reduce the argument, report any loss of precision, and
	   note in which octant x (mod two_pi) lay.		  */

    octant = reduceArg( x ) ;
    if( dp_lost > 0 )  {
	fprintf( stderr, 
	"apmSin : Big argument, lost %d decimal places from the answer. \n", 
								  dp_lost ) ;
	apm_errno = APM_WTRUNC ; 
    }
    else
	apm_errno = APM_OK ;

	/* Evaluate the sine.  Which of the two reduced functions
	   one uses depends on the octant.			*/

    switch( octant ) {
	case 0 :
	    reducedSin() ;
	    break ;

	case 1 :
	    reducedCos() ;
	    break ;

	case 2 :
	    reducedCos() ;
	    break ;

	case 3 :
	    reducedSin() ;
	    break ;

	case 4 :
	    reducedSin() ;
	    apmNegate( scratch, Answer ) ;
	    apmAssign( Answer, scratch ) ;
	    break ;

	case 5 :
	    reducedCos() ;
	    apmNegate( scratch, Answer ) ;
	    apmAssign( Answer, scratch ) ;
	    break ;

	case 6 :
	    reducedCos() ;
	    apmNegate( scratch, Answer ) ;
	    apmAssign( Answer, scratch ) ;
	    break ;

	case 7 :
	    reducedSin() ;
	    apmNegate( scratch, Answer ) ;
	    apmAssign( Answer, scratch ) ;
	    break ;

	default :
	    break ;
    }

    apmAssign( result, Answer ) ;

# if DEBUG
    fprintf( stderr, "apmSin( %.4f ) : octant = %d, answer = %.4f \n",
			   apmtodbl( x ), octant, apmtodbl( Answer ) ) ;
# endif
    return ;
}
/* +++++++++++++++++++++++++ */


void 
apmCos (APM result, APM x)
{
    int   octant ;

    if( specialsInit == NO ) {
	fprintf( stderr, 
		 "apmCos() : Please call initApmSpecials() first. \n" ) ;
 	apmAssignLong( result, 0L, 0, 0 ) ;
	apm_errno = APM_EPARM ;
	return ;
    }
    else 
	apm_errno = APM_OK ;

	/* Reduce the argument, report any loss of precision, and
	   note in which octant x (mod two_pi) lay.		  */


    octant = reduceArg( x ) ;
    if( dp_lost > 0 )  {
	fprintf( stderr, 
	"apmCos : Big argument, lost %d decimal places from the answer. \n", 
								  dp_lost ) ;
	apm_errno = APM_WTRUNC ; 
    }
    else
	apm_errno = APM_OK ;

	/* Evaluate the cosine.  Which of the two reduced functions
	   one uses depends on the octant.			  */

    switch( octant ) {
	case 0 :
	    reducedCos() ;
	    break ;

	case 1 :
	    reducedSin() ;
	    break ;

	case 2 :
	    reducedSin() ;
	    apmNegate( scratch, Answer ) ;
	    apmAssign( Answer, scratch ) ;
	    break ;

	case 3 :
	    reducedCos() ;
	    apmNegate( scratch, Answer ) ;
	    apmAssign( Answer, scratch ) ;
	    break ;

	case 4 :
	    reducedCos() ;
	    apmNegate( scratch, Answer ) ;
	    apmAssign( Answer, scratch ) ;
	    break ;

	case 5 :
	    reducedSin() ;
	    apmNegate( scratch, Answer ) ;
	    apmAssign( Answer, scratch ) ;
	    break ;

	case 6 :
	    reducedSin() ;
	    break ;

	case 7 :
	    reducedCos() ;
	    break ;

	default :
	    break ;
    }

    apmAssign( result, Answer ) ;
    return ;
}
/* +++++++++++++++++++++++++ */

void 
apmSqrt ( 
/*
    Find square roots using Newton's method.
						*/
    APM result,
    int dp,
    APM x
)
{
    int     comp, dp_plus ;
    APM     *this_approx, *next_approx, *temp ;

/*
	Check that all the scratch variables are ready.
*/

    if( specialsInit == NO ) {
	fprintf( stderr, 
		 "apmSqrt() : Please call initApmSpecials() first. \n" ) ;
 	apmAssignLong( result, 0L, 0, 0 ) ;
	apm_errno = APM_EPARM ;
	return ;
    }
    else 
	apm_errno = APM_OK ;

/*
	If the argument is zero, just return zero.
	If the argument is negative, whine.
*/
    if( (comp = apmCompare( x, zero )) == 0 ) {
	apmAssign( result, zero ) ;
	return ;
    }
    else if( comp == -1 ) {
	fprintf( stderr, "apmSqrt() : Can't handle negative arguments.\n" ) ;
	apm_errno = APM_EPARM ;
	return ;
    }
    else
	apm_errno = APM_OK ;

/*
	Do up Newton.  The rule is

		y[n+1] = (y[n] + x/y[n]) / 2.0
*/
    dp_plus = dp + 2 ;
    apmAssignLong( ub_diff, 1L, -dp_plus, 0 ) ;

    this_approx = &approx[0] ;
    next_approx = &approx[1] ;

    apmAssign( *this_approx, x ) ;
    apmAssign( *next_approx, zero ) ;
    apmSubtract( diff, *this_approx, *next_approx ) ;
    while( apmCompare( diff, ub_diff) > 0 ) {
	apmDivide( scratch, dp_plus, (APM) NULL, x, *this_approx ) ;
	apmCalc( scratch, scratch, *this_approx, APM_ADD, NULL ) ;
	apmDivide( *next_approx, dp_plus, (APM) NULL, scratch, two ) ;
	apmTruncate( *next_approx, dp_plus ) ; 

	apmCalc( diff, *this_approx, *next_approx, APM_SUB, APM_ABS, NULL ) ;
	m_swap( this_approx, next_approx, temp ) ;
    }

    apmAssign( result, *this_approx ) ;
    return ;
}
/* +++++++++++++++++++++++++++++++++++++++ */

void 
apmFloor (APM result, APM arg, int base)
{
    char  buf[BUF_SZ], *cpt ;

    apmConvert( buf, BUF_SZ, 2, NO_ROUND, LEFT_JUST, arg ) ;
    for( cpt = buf ; *cpt != '\0' ; cpt++ )
	if( *cpt == '.' )
	    *cpt = '\0' ;

    apmAssignString( result, buf, base ) ;
}
/* ++++++++++++++++++++++++++++++++ */

void setTrigCoef( void ) 
{
   int   j, order, coef_num ;

   sinCoef = (APM *) malloc( trig_terms * sizeof( APM ) ) ;
   cosCoef = (APM *) malloc( trig_terms * sizeof( APM ) ) ;
   if( (sinCoef == NULL) || (cosCoef == NULL) ) {
       fprintf( stderr, "Trouble allocating %d APMs for coefficients.\n", trig_terms ) ;
       exit(0) ;
    }

   for( j=0 ; j < trig_terms ; j++ ) {
       sinCoef[j] = apmNew( 0 ) ;
       cosCoef[j] = apmNew( 0 ) ;
    }

    if( (trig_terms % 2) != 0 ) {
	apmAssignLong( sinCoef[trig_terms-1], -1L, 0, 0 ) ;
	apmAssignLong( cosCoef[trig_terms-1], -1L, 0, 0 ) ;
    }
    else {
	apmAssignLong( sinCoef[trig_terms-1], 1L, 0, 0 ) ;
	apmAssignLong( cosCoef[trig_terms-1], 1L, 0, 0 ) ;
    }

    coef_num = trig_terms - 2 ;
    for( order = (2 * trig_terms - 1) ; order > 1 ; order -= 2 ) {
	/* coefficients for the sine */

	apmAssignLong( apmOrder, -((long) order), 0, 0 ) ;
	apmMultiply( scratch, sinCoef[coef_num+1], apmOrder ) ;
	apmAssignLong( apmOrder, (long)(order-1), 0, 0 ) ;
	apmMultiply( sinCoef[coef_num], scratch, apmOrder ) ;

	/* coefficients for the cosine */
	apmMultiply( scratch, cosCoef[coef_num+1], apmOrder ) ;
	apmAssignLong( apmOrder, -(long)(order-2), 0, 0 ) ;
	apmMultiply( cosCoef[coef_num], scratch, apmOrder ) ;

	coef_num-- ;
    }

    apmAssign( sinFactrl, sinCoef[0] ) ;
    apmAssign( cosFactrl, cosCoef[0] ) ;
}
/* +++++++++++++++++++++++++++++++++++++++++++++++ */

void 
apmFloorString (char *s, int n, APM x)
{
    apmConvert( s, n, 1, NO_ROUND, LEFT_JUST, x ) ;
    strip_frac( s ) ;
}
/* +++++++++++++++++++++ */

void 
strip_frac (char *str)
{
    char  *cpt ;

    for( cpt = str ; cpt != '\0' ; cpt++ )
	if( *cpt == '.' ) {
	    *cpt = '\0' ;
	    break ;
	}
}
/* +++++++++++++++++++++++ */

int 
apmLogBd (APM x)
/*
	Computes an upper bound on the base-10 log of the 
	fractional part of an apm.
*/
{
    int   order ;
    char  *bpt ;

    if( apmCompare( one, x ) <= 0 ) {
		apmFloorString( log_buf, BUF_SZ, x ) ;
		return( strlen( log_buf ) ) ;
    }
    else {
    	/* Convert x into a string */
		apmConvert( log_buf, BUF_SZ, (BUF_SZ-4), NO_ROUND, LEFT_JUST, x ) ;
		
		/* Get a pointer to the first digit to the right of the decimal point */
		for( bpt=log_buf ; *bpt != '.' ; bpt++ ) ;
		bpt++ ;
		
		/* Count the number of zeroes to the right of the decimal point. */
		for( order=0 ; *bpt == '0' ; bpt++, order-- ) ;
		return( order ) ;
    }
}

