# include <stdio.h>
# include <math.h>

# include "apm.h"
# include "apmUtil.h"
# include "converse.h"
# include "pi.h"

# define  DEBUG	 	NO 

# define  DELTA		0.01
# define  MAX_JUMP	0.1

/* A variable shared via converse.h */
APM  Rstart_size ;
/* ++++++++++++++++++++++++++++++++++++++ */

void 
setHermStart (RPrism *priz)
{
    double  a, b, c, two_c, x, y ;
    double  jump_sz, jump_scl, dx, dy ;
    double  gx, gy, hxx, hxy, hyy, hdet, tolerance ;

    a = apmtodbl( priz->center->p[0] ) ;
    b = apmtodbl( priz->center->p[1] ) ;
    c = apmtodbl( priz->center->p[2] ) ;
    two_c = 2.0 * c ;

# if DEBUG
    printf( "a = %.6e, b= %.6e, c= %.6e \n", a, b, c ) ;
    fflush( stdout ) ;
# endif

    tolerance = NEWT_TOL * (fabs(a) + fabs(b) + fabs(c)) ;

/*   
	Use Newton's method to try to find a minimum for the
	trace of the matrix beta.
*/

    x = HALF_PI ;
    y = HALF_PI ;

    do {
	/*	components of the gradient.	*/

	gx =  -a * cos( x )  -  two_c * cos( x + y ) ;
	gy =  -b * cos( y )  -  two_c * cos( x + y ) ;

	/* 	components of the Hessian 	*/

	hxx =  a * sin( x )  +  two_c * sin( x + y ) ;
	hxy =  two_c * sin( x + y ) ;
	hyy =  b * sin( y )  +  two_c * sin( x + y ) ;
	hdet = hxx * hyy  -  hxy * hxy ;

	/*  	A Newton's method step		*/
	if( hdet != 0.0 ) {
	    dx = (  gx * hyy  -  gy * hxy ) / hdet ;
	    dy = ( -gx * hxy  +  gy * hxx ) / hdet ;

	    if( (jump_sz = fabs(dx) + fabs(dy)) > MAX_JUMP ) {
		jump_scl = MAX_JUMP / jump_sz ;
		dx *= jump_scl ;
		dy *= jump_scl ;
	    }

	    x -= dx ;
	    y -= dy ;
		
# if DEBUG
	printf( "mid - Newton says : x = %.6e, y= %.6e \n", x, y ) ;
	fflush( stdout ) ;
# endif
	}
	else {
	    fprintf( stderr, "Death during Newton's method. \n" ) ;
	    cease() ;
	}
    } while( (fabs(gx) + fabs(gy))  > tolerance ) ;	

/*
	Force the starting point to lie on the line x=y.
*/
    dbltoapm( priz->center->z.u[0], BASE, x ) ;
    dbltoapm( priz->center->z.u[1], BASE, x ) ;

#if DEBUG
    printf( "Herman's starting point : x = %.6e, y= %.6e \n", x, x ) ;
    fflush( stdout ) ;
# endif
}
/* ++++++++++++++++++++++ */

void 
setLLStart (RPrism *priz)
{
/*
	Beware : this function expects to be called AFTER
	setHermStart(), no matter which criterion is in force.
*/
    double  discrim, sqrt_disc, sqrt() ;
    double  a_sin, a_cos, b_sin, b_cos, c_sin, c_cos ;
    double  a, b, c, two_c, x, y ;
    double  jump_sz, jump_scl, dx, dy ;
    double  gx, gy, hxx, hxy, hyy, hdet, tolerance ;
    double  dDisc_dx, dDisc_dy ;

    a = apmtodbl( priz->center->p[0] ) ;
    b = apmtodbl( priz->center->p[1] ) ;
    c = apmtodbl( priz->center->p[2] ) ;
    two_c = 2.0 * c ;

    x = apmtodbl( priz->center->z.u[0] ) ;
    y = apmtodbl( priz->center->z.u[1] ) ;

# if DEBUG
    printf( "setLLStart : a = %.6e, b= %.6e, c= %.6e \n", a, b, c ) ;
    printf( "setLLStart : x = %.6f,  y = %.6f \n", x, y ) ;
    fflush( stdout ) ;
# endif

    tolerance = NEWT_TOL * (a + b + c) ;


    do {
	/*	preliminaries			*/

	a_sin = a * sin( x ) ;
	b_sin = b * sin( y ) ;
	c_sin = two_c * sin( x + y ) ;

	a_cos = a * cos( x ) ;
	b_cos = b * cos( y ) ;
	c_cos = two_c * cos( x + y ) ;

	discrim = ( a_sin - b_sin ) * ( a_sin - b_sin ) +
					  c_sin * c_sin ;
	sqrt_disc = sqrt( discrim ) ;
	dDisc_dx = a_cos * (a_sin - b_sin) + c_cos * c_sin ;
	dDisc_dy = b_cos * (b_sin - a_sin) + c_cos * c_sin ;

# if DEBUG
	printf( "mid-Newton, lambda = %.6f \n", 
		(4.0 - a_sin - b_sin - c_sin - sqrt_disc) / 2.0 ) ;
# endif

	/*	components of the gradient.	*/

	gx = -a_cos - c_cos - dDisc_dx / sqrt_disc ;
	gy = -b_cos - c_cos - dDisc_dy / sqrt_disc ;

	/* 	components of the Hessian 	*/

	hxx = a_sin + c_sin +
	      ( a_sin * (a_sin - b_sin) - 
		a_cos * a_cos - c_cos * c_cos + 
		c_sin * c_sin ) / sqrt_disc
	      + dDisc_dx * dDisc_dx / (discrim * sqrt_disc) ;

	hxy = c_sin +
	      ( a_cos * b_cos + c_sin * c_sin -
				c_cos * c_cos ) / sqrt_disc
	      + dDisc_dx * dDisc_dy / (discrim * sqrt_disc) ;

	hyy = b_sin + c_sin +
	      ( b_sin * (b_sin - a_sin) - 
		b_cos * b_cos - c_cos * c_cos + 
		c_sin * c_sin ) / sqrt_disc
	      + dDisc_dy * dDisc_dy / (discrim * sqrt_disc) ;

	hdet = hxx * hyy  -  hxy * hxy ;

	/*  	A Newton's method step		*/
	if( hdet != 0.0 ) {
	    dx = (  gx * hyy  -  gy * hxy ) / hdet ;
	    dy = ( -gx * hxy  +  gy * hxx ) / hdet ;

	    if( (jump_sz = fabs(dx) + fabs(dy)) > MAX_JUMP ) {
		jump_scl = MAX_JUMP / jump_sz ;
		dx *= jump_scl ;
		dy *= jump_scl ;
	    }

	    x -= dx ;
	    y -= dy ;
		
# if DEBUG
	printf( "mid - Newton says : x = %.6e, y= %.6e \n", x, y ) ;
	fflush( stdout ) ;
# endif
	}
	else {
	    fprintf( stderr, "Death during Newton's method. \n" ) ;
	    cease() ;
	}
    } while( (fabs(gx) + fabs(gy))  > tolerance ) ;	

/*
	Force the starting point to lie on the line x=y.
*/
    dbltoapm( priz->center->z.u[0], BASE, x ) ;
    dbltoapm( priz->center->z.u[1], BASE, x ) ;

#if DEBUG
    printf( "Least eigenvalue starting point : x = %.6e, y= %.6e \n", x, x ) ;
    fflush( stdout ) ;
# endif
}
/* +++++++++++++++++++++ */

void 
shiftStart (
/* 
    Shift the starting point off the main diagonal. 
*/
    RPrism *priz
)
{
    double    x, y, a, b, amin, bmin ;

    a = apmtodbl( priz->center->p[0] ) ;
    b = apmtodbl( priz->center->p[1] ) ;

    amin = a - apmtodbl( priz->matrix[0] ) ;
    bmin = b - apmtodbl( priz->matrix[MAT_DIM+1] ) ;

    x = apmtodbl( priz->center->z.u[0] ) ;
    y = apmtodbl( priz->center->z.u[1] ) ;

    if( fabs(x - y) < DELTA ) {
        if( amin < bmin ) {
	    x += DELTA ;
	    y -= DELTA ;
	}
    	else {
	    x -= DELTA ;
	    y += DELTA ;
	}
    }

    dbltoapm( priz->center->z.u[0], BASE, x ) ;
    dbltoapm( priz->center->z.u[1], BASE, y ) ;
}
